import gzip
from timeit import timeit
import unittest
import zlib_state

class TestZlibState(unittest.TestCase):
    def test_GzipStateFile_fullextract(self):
        with zlib_state.GzipStateFile('testdata/frankenstein.txt.gz') as f:
            all_text = f.read()
        with gzip.open('testdata/frankenstein.txt.gz', 'rb') as f:
            source_text = f.read()
        self.assertEqual(all_text, source_text)

    def test_compare_times(self):
        def gzip_run():
            with gzip.open('testdata/frankenstein.txt.gz', 'rb') as f:
                all_text = f.read()
        def gzip_state_run():
            with zlib_state.GzipStateFile('testdata/frankenstein.txt.gz') as f:
                all_text = f.read()
        def gzip_state_run_keep_last():
            with zlib_state.GzipStateFile('testdata/frankenstein.txt.gz', keep_last_state=True) as f:
                all_text = f.read()
        def zlib_state_run():
            decomp = zlib_state.Decompressor(32 + 15)
            with open('testdata/frankenstein.txt.gz', 'rb') as f:
                while not decomp.eof():
                    needed_input = decomp.needs_input()
                    if needed_input > 0:
                        decomp.feed_input(f.read(needed_input))
                    decomp.read()
        # Warm up cache
        gzip_run()
        gzip_state_run()
        gzip_state_run_keep_last()
        zlib_state_run()
        print("gzip_run", timeit(gzip_run, number=100), "s/100")
        print("gzip_state_run", timeit(gzip_state_run, number=100), "s/100")
        print("gzip_state_run_keep_last", timeit(gzip_state_run_keep_last, number=100), "s/100")
        print("zlib_state_run", timeit(zlib_state_run, number=100), "s/100")
        # It doesn't look like my buffered implementation is quite as fast as gzip's
        # Keeping the last sate also hurts a bit. But when not using the buffered version, it's faster!
        # gzip_run 0.23395611305022612 s/100
        # gzip_state_run 0.294685336004477 s/100
        # gzip_state_run_keep_last 0.316738452995196 s/100
        # zlib_state_run 0.1936574190040119 s/100

    def test_GzipStateFile_zseek(self):
        TARGET_LINE = 5000 # pick back up after around the 5,000th line
        with zlib_state.GzipStateFile('testdata/frankenstein.txt.gz', keep_last_state=True) as f:
            for i, line in enumerate(f):
                if i == TARGET_LINE:
                    state, pos = f.last_state, f.last_state_pos

        with zlib_state.GzipStateFile('testdata/frankenstein.txt.gz', keep_last_state=True) as f:
            f.zseek(pos, state)
            remainder = f.read()
            # The 5,000th line should be there
            self.assertTrue(b'day.  He pointed out to me the shifting colours of the landscape and' in remainder)
            self.assertEqual(len(remainder), 207371)


if __name__ == '__main__':
    unittest.main()



import io
import os
if hasattr(os, 'add_dll_directory'):
    # On windows, need to be sure that zlib is in the dll directory path before loading ext module
    # In particular, needed by python 3.8+ on Windows
    with os.add_dll_directory(os.environ.get("ZLIB_HOME", "C:/Program Files/zlib/bin")):
        from _zlib_state import Decompressor
else:
    from _zlib_state import Decompressor


class GzipStateFileBase(io.RawIOBase):
    def __init__(self, path, keep_last_state=False):
        if isinstance(path, str):
            self.file = open(path, 'rb')
        else:
            self.file = path
        self.decomp = Decompressor(32 + 15)
        self.last_state = None
        self.last_state_pos = None
        self.keep_last_state = keep_last_state

    def zseek(self, pos, state):
        self.file.seek(pos)
        self.decomp = Decompressor(-15)
        self.decomp.set_state(state[0], state[1], state[2])

    def readinto(self, buf):
        if len(buf) == 1:
            raise ValueError()
        count = 0
        while count == 0 and not self.decomp.eof():
            needed_bytes = self.decomp.needs_input()
            if needed_bytes > 0:
                # decompressor needs more input
                self.decomp.feed_input(self.file.read(needed_bytes))
            count += self.decomp.read(outbytes=buf)
            if self.keep_last_state and self.decomp.block_boundary():
                self.last_state = self.decomp.get_state()
                self.last_state_pos = self.decomp.total_in()
        return count

    def read(self, size=-1):
        if size == 1:
            raise ValueError()
        if size == -1:
            return super().read(size) # reads entire file
        result = None
        while not result and not self.decomp.eof():
            needed_bytes = self.decomp.needs_input()
            if needed_bytes > 0:
                # decompressor needs more input
                self.decomp.feed_input(self.file.read(needed_bytes))
            result = self.decomp.read(outsize=size)
            if self.decomp.block_boundary():
                self.last_state = self.decomp.get_state()
                self.last_state_pos = self.decomp.total_in()
        return result

    def eof(self):
        return self.decomp.eof()

    def __bool__(self):
        return not self.decomp.eof()

    def readable(self):
        return not self.decomp.eof()

    def fileno(self):
        return self.file.fileno()

    def seek(self):
        raise NotImplementedError()

    def truncate(self):
        raise NotImplementedError()

    def close(self):
        super().close()
        self.file.close()


class GzipStateFile(io.BufferedIOBase):
    def __init__(self, path, keep_last_state=False, buffer_size=io.DEFAULT_BUFFER_SIZE):
        self.raw = GzipStateFileBase(path, keep_last_state)
        self.buffer_size = buffer_size
        self.buffer = bytearray(buffer_size)
        self.buffer_start = 0
        self.buffer_stop = 0

    def detach(self):
        return self.raw

    def read(self, count=-1):
        if count == -1:
            return self._readall()
        result = b''
        while len(result) < count and not self.raw.eof():
            result += self.read1(count - len(result))
        return result

    def _readall(self):
        result = b''
        while not self.raw.eof():
            result += self.read(self.buffer_size)
        return result

    def read1(self, count=-1):
        start, stop = self.buffer_start, self.buffer_stop
        if start == stop:
            # buffer empty
            size = self.raw.readinto(self.buffer)
            self.buffer_start, self.buffer_stop = start, stop = 0, size
        size = stop - start
        if count != -1: # indicates to read as much as possible
            size = min(size, count)
        result = self.buffer[start:start+size]
        self.buffer_start += size
        return result

    def readinto(self, buf):
        count = 0
        while count < len(buf) and not self.raw.eof():
            count += self.readinto1(buf[count:])
        return count

    def readinto1(self, buf):
        start, stop = self.buffer_start, self.buffer_stop
        if start == stop:
            # buffer empty
            size = self.raw.readinto(self.buffer)
            self.buffer_start, self.buffer_stop = start, stop = 0, size
        size = min(stop - start, count)
        buf[:size] = self.buffer[start:start+size]
        self.buffer_start += size
        return size

    def readline(self):
        if self.buffer_start != self.buffer_stop:
            try:
                idx = self.buffer[self.buffer_start:self.buffer_stop].index(b'\n')
                # line found in buffer. Use that and advance buffer pointers
                line = self.buffer[self.buffer_start:self.buffer_start+idx+1]
                self.buffer_start += idx + 1
                return line
            except ValueError:
                # line break not found in remaining buffer. Start with the remaining buffer and carry on
                line = self.buffer[self.buffer_start:self.buffer_stop]
                self.buffer_start, self.buffer_stop = 0, 0 # buffer consumed
        else:
            line = b''
        while not line.endswith(b'\n') and not self.raw.eof():
            chunk = self.read1(self.buffer_size)
            try:
                idx = chunk.index(b'\n')
                self.buffer_start -= (len(chunk) - idx - 1) # move back the buffer cursor to the end of the line
                chunk = chunk[:idx+1]
            except ValueError:
                pass # \n not found, that's OK
            line += chunk
        return line

    def peek(self, count=-1):
        start, stop = self.buffer_start, self.buffer_stop
        if start == stop:
            # buffer empty
            size = self.raw.readinto(self.buffer)
            self.buffer_start, self.buffer_stop = start, stop = 0, size
        size = stop - start
        if count != -1: # indicates to read as much as possible
            size = min(size, count)
        result = self.buffer[start:start+size]
        # do NOT advance the start pointer
        return result

    def zseek(self, pos, state):
        self.raw.zseek(pos, state)
        self.buffer_start, self.buffer_stop = 0, 0

    def eof(self):
        return self.raw.eof()

    def __getattr__(self, key):
        if key in ('last_state', 'last_state_pos'):
            return getattr(self.raw, key)
        return getattr(super(), key)

    def close(self):
        super().close()
        self.raw.close()
        self.buffer = None



import os
import sys
from glob import glob
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

include_dirs, libs, lib_dirs = [], [], []

windows = sys.platform.startswith("win")

if windows:
    ZLIB_HOME = os.environ.get("ZLIB_HOME", "C:/Program Files/zlib")
    include_dirs.append(os.path.join(ZLIB_HOME, "include"))
    libs.append('zlib')
    lib_dirs.append(os.path.join(ZLIB_HOME, "lib"))
else:
    libs.append('z')

setuptools.setup(
    name="zlib-state",
    version="0.0.1",
    author="Sean MacAvaney",
    author_email="sean.macavaney@gmail.com",
    description="Low-level interface to the zlib library that enables capturing the decoding state",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seanmacavaney/zlib-state",
    include_package_data = True,
    packages=setuptools.find_packages(include=['zlib_state']),
    install_requires=[],
    classifiers=[
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.5',
    ext_modules=[setuptools.Extension("_zlib_state", ["src/zlib_state.c"], libraries=libs, library_dirs=lib_dirs, include_dirs=include_dirs)],
)
